"""Common type definitions."""
from __future__ import annotations

FEATURE = "feature"
SCENARIO_OUTLINE = "scenario outline"
EXAMPLES = "examples"
EXAMPLES_HEADERS = "example headers"
EXAMPLE_LINE = "example line"
SCENARIO = "scenario"
BACKGROUND = "background"
GIVEN = "given"
WHEN = "when"
THEN = "then"
TAG = "tag"

STEP_TYPES = (GIVEN, WHEN, THEN)
